/********************************************************************
 ********************************************************************
  Interface for MATLAB and SIPAMPL version 2

  A. Ismael F. Vaz
  Edite M. G. P. Fernandes

  University of Minho, Portugal

  11/11/2002 Updated


  This files provides the function sipampl2 to MATLAB.
  
    ...=sipampl2(opt,...)
    
  sipampl2 will behave differently for each opt value.




*********************************************************************
*********************************************************************/
#include <sys/stat.h>
#include "mex.h"
#include "matrix.h"
#undef printf
#include "../../asl_pfgh.h"
#include "../sip.h"


static char msgbuf[256];
ASL *asl=NULL;
static int neq=-1, nineq=-1;


static void clean_house(void) {


  if (asl)
    ASL_free(&asl);
  asl=NULL;

  sip_free();

  nineq=-1;neq=-1;
}


static void usage(void) {
 mexPrintf("sipampl2 is a MATLAB function to be called by the sip_xxx functions\n");
 mexPrintf("If you observed this error using a function sip_xxx please report the\n");
 mexPrintf("  bug to aivaz@dps.uminho.pt, otherwise look into the sip_xxx.m file to\n");
 mexPrintf("  find how to use sipampl2\n");
}


void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
  FILE *nl;
  char stub[512];
  char stubnl[512];
  fint m, n, no, nz, mxr, mxc;
  real *X, *T, *RX, *RT, *RXT, *nn, *C, *CEQ, **KR, *S, *opt;
  int i, j, k, neq_tmp, nineq_tmp;
  int mx, nx, mt, nt;
#if defined __linux__ || defined linux
  struct stat stat_file;
#else
  struct _stat stat_file;
#endif


  if(nrhs<1 || !mxIsDouble(prhs[0])){
    usage();
    return;
  }

  opt=mxGetPr(prhs[0]);

  if(*opt && !asl)
    mxErrMsgTxt("You must call\n\t[x0,xbl,xbu,tbl,tbu]=sipampl2(0,stub)\nor\n"
        "\t[x0,xbl,xbu,tbl,tbu]=sip_init(stub)\n first");

  switch((int)*opt){
    case 0:
        /* sip_init */
        if(nrhs<2 || !mxIsChar(prhs[1]) || nlhs!=5){
            usage();
            return;
        }
        if(mxGetString(prhs[1], stub, sizeof(stub)))
            mexErrMsgTxt("First argument must be 'stub'\n");
        clean_house();

        /* Check if file 'stub.nl' exists */
        sprintf(stubnl, "%s.nl", stub);
#if defined __linux__ && defined linux
        if(stat(stubnl, &stat_file)){
#else
        if(_stat(stubnl, &stat_file)){
#endif
            mexPrintf("File %s.nl does not exist\n", stub);
	    mexErrMsgTxt("Problem does not exist\n");
            return;
        }

        asl = ASL_alloc(ASL_read_pfgh);
        want_xpi0=1;

        mexAtExit(clean_house);



        nl=jac_dim_ASL(asl, stub, &m, &n, &no, &nz, &mxr, &mxc, (fint)strlen(stub));
        if (!nl)
            mexErrMsgTxt("Can't read problem\n");

        pfgh_read_ASL(asl, nl,  0);



        if (mxr == 0 || mxc == 0)
            mxErrMsgTxt("Semi-Infinite Programming interface needs .row and .col files.\n"
                "In order to provide .row and .col files, in your AMPL session say\n"
                "\toption auxfiles rc;\nbefore saying\n\twrite ...\n");

        sip_init(asl);

        if(ntsip<1 || ntsip>2)
            mxErrMsgTxt("Number of semi-infinite variables must be positive and less or equal to 2\n");
        if(ntsipc<1)
            mxErrMsgTxt("Number of semi-infinite constraints must be greater then zero.\n"
            "Semi-Infinite Programming interface needs .row and .col files.\n"
            "In order to provide .row and .col files, in your AMPL session say\n"
            "\toption auxfiles rc;\nbefore saying\n\twrite ...\n");


        /* nlhs is 5 */
        if(X0 && (!havex0 || (havex0 && *havex0))){
            X=mxGetPr(plhs[0]=mxCreateDoubleMatrix(nxsip, 1, mxREAL));
            sip_extractx(X0, X);
        } else {
            X=mxGetPr(plhs[0]=mxCreateDoubleMatrix(1, 1, mxREAL));
            *X=mxGetNaN();
        }
        

        nn=mxGetPr(plhs[1]=mxCreateDoubleMatrix(nxsip, 1, mxREAL));
        memcpy(nn,XBL,nxsip*sizeof(double));

        nn=mxGetPr(plhs[2]=mxCreateDoubleMatrix(nxsip, 1, mxREAL));
        memcpy(nn,XBU,nxsip*sizeof(double));

        nn=mxGetPr(plhs[3]=mxCreateDoubleMatrix(ntsip, 1, mxREAL));
        memcpy(nn,TBL,ntsip*sizeof(double));

        nn=mxGetPr(plhs[4]=mxCreateDoubleMatrix(ntsip, 1, mxREAL));
        memcpy(nn,TBU,ntsip*sizeof(double));

        break;
    case 1:
        /* sip_objval */
        if (nlhs!=1 || nrhs!=2 || !mxIsDouble(prhs[1])){
            usage();
            return;
        }
        
        mx=mxGetM(prhs[1]);
        nx=mxGetN(prhs[1]);
        if((mx!=1 || nx!=nxsip) && (mx!=nxsip || nx!=1)) {
            mexPrintf("%s must be a vector of %d elements\n", mxGetName(prhs[1]), nxsip);
            usage();
            return;
        }
        
        X=mxGetPr(prhs[1]);
        nn=mxGetPr(plhs[0]=mxCreateDoubleMatrix(1, 1, mxREAL));
        nn[0]=sip_objval(0, X, NULL);        
        break;
    case 2:
        /* sip_objgrd */
        if (nlhs!=1 || nrhs!=2 || !mxIsDouble(prhs[1])){
            usage();
            return;
        }
        
        mx=mxGetM(prhs[1]);
        nx=mxGetN(prhs[1]);
        if((mx!=1 || nx!=nxsip) && (mx!=nxsip || nx!=1)) {
            mexPrintf("%s must be a vector of %d elements\n", mxGetName(prhs[1]), nxsip);
            usage();
            return;
        }
        
        X=mxGetPr(prhs[1]);
        RX=mxGetPr(plhs[0]=mxCreateDoubleMatrix(nxsip, 1, mxREAL));
        sip_objgrd(0, X, RX, NULL);
        break;
    case 3:
        /* sip_objhes */
        if (nlhs!=1 || nrhs!=1){
            usage();
            return;
        }
        RX=mxGetPr(plhs[0]=mxCreateDoubleMatrix(nxsip, nxsip, mxREAL));
        sip_objhes(0, RX, NULL);
        break;
    case 4:
        /* sip_conval */
        if(nlhs!=3 || nrhs!=3 || !mxIsDouble(prhs[1]) || !mxIsDouble(prhs[2])){
            usage();
            return;
        }
        X=mxGetPr(prhs[1]);
        T=mxGetPr(prhs[2]);

        mx=mxGetM(prhs[1]);
        nx=mxGetN(prhs[1]);
        mt=mxGetM(prhs[2]);
        nt=mxGetN(prhs[2]);
        if((mx!=1 || nx!=nxsip) && (mx!=nxsip || nx!=1)) {
            mexPrintf("%s must be a vector of %d elements\n", mxGetName(prhs[1]), nxsip);
            usage();
            return;
        }
        if((mt!=1 || nt!=ntsip) && (mt!=ntsip || nt!=1)) {
            mexPrintf("%s must be a vector of %d elements\n", mxGetName(prhs[2]), ntsip);
            usage();
            return;
        }
        
        if(neq==-1 || nineq==-1){
            /* count how many equality constraints */
            for(i=0,neq=0;i<nxsipc;i++)
                if(XCBL[i]==XCBU[i])neq++;
            nineq=nxsipc-neq;
        }

        C=mxGetPr(plhs[0]=mxCreateDoubleMatrix(nineq, 1, mxREAL));
        CEQ=mxGetPr(plhs[1]=mxCreateDoubleMatrix(neq, 1, mxREAL));
        RT=mxGetPr(plhs[2]=mxCreateDoubleMatrix(ntsipc, 1, mxREAL));
        RX=mymalloc(nxsipc*sizeof(double));
        
        sip_conval(X, T, RX, RT, NULL);
        for(i=0,neq_tmp=0,nineq_tmp=0;i<nxsipc;i++){
            if(XCBL[i]!=XCBU[i]){
                C[nineq_tmp++]=RX[i]-XCBU[i]; /* Only lower or equal finite constraints */
            } else {
                CEQ[neq_tmp++]=RX[i]-XCBU[i];
            }
        }
        for(i=0;i<ntsipc;i++)
          RT[i]-=TCBU[i]; /* Only lower or equal infinite constraints */
        free(RX);
        break;
    case 5:
        /* sip_conxeqval */
        if(nlhs!=1 || nrhs!=3 || !mxIsDouble(prhs[1]) || !mxIsDouble(prhs[2])){
            usage();
            return;
        }
        
        nn=mxGetPr(prhs[1]);
        X=mxGetPr(prhs[2]);

        mx=mxGetM(prhs[2]);
        nx=mxGetN(prhs[2]);
        if((mx!=1 || nx!=nxsip) && (mx!=nxsip || nx!=1)) {
            mexPrintf("%s must be a vector of %d elements\n", mxGetName(prhs[2]), nxsip);
            usage();
            return;
        }

        CEQ=mxGetPr(plhs[0]=mxCreateDoubleMatrix(1, 1, mxREAL));
        
        for(i=0,neq_tmp=0;i<nxsipc;i++){
            if(XCBL[i]==XCBU[i]){
                /* equality found */
                if(neq_tmp==(int)nn[0])break;
                neq_tmp++;
            }
        }
        if(i==nxsipc){
            mxErrMsgTxt("Invalid constraint number\n");
        } else {
            CEQ[0]=sip_conxval(i, X, NULL)-XCBU[i];
        }
        break;
    case 6:
        /* sip_conxineqval */
        if(nlhs!=1 || nrhs!=3 || !mxIsDouble(prhs[1]) || !mxIsDouble(prhs[2])){
            usage();
            return;
        }
        nn=mxGetPr(prhs[1]);
        X=mxGetPr(prhs[2]);

        mx=mxGetM(prhs[2]);
        nx=mxGetN(prhs[2]);
        if((mx!=1 || nx!=nxsip) && (mx!=nxsip || nx!=1)) {
            mexPrintf("%s must be a vector of %d elements\n", mxGetName(prhs[2]), nxsip);
            usage();
            return;
        }

        C=mxGetPr(plhs[0]=mxCreateDoubleMatrix(1, 1, mxREAL));
        
        for(i=0,nineq_tmp=0;i<nxsipc;i++){
            if(XCBL[i]!=XCBU[i]){
                /* inequality found */
                if(nineq_tmp==(int)nn[0])break;
                nineq_tmp++;
            }
        }
        if(i==nxsipc){
            mxErrMsgTxt("Invalid constraint number\n");
        } else {
            C[0]=sip_conxval(i, X, NULL)-XCBU[i];
        }
        break;
    case 7:
        /* sip_conxeqgrd */
        if(nlhs!=1 || nrhs!=3 || !mxIsDouble(prhs[1]) || !mxIsDouble(prhs[2])){
            usage();
            return;
        }
        nn=mxGetPr(prhs[1]);
        X=mxGetPr(prhs[2]);

        mx=mxGetM(prhs[2]);
        nx=mxGetN(prhs[2]);
        if((mx!=1 || nx!=nxsip) && (mx!=nxsip || nx!=1)) {
            mexPrintf("%s must be a vector of %d elements\n", mxGetName(prhs[2]), nxsip);
            usage();
            return;
        }

        CEQ=mxGetPr(plhs[0]=mxCreateDoubleMatrix(nxsip, 1, mxREAL));
        
        for(i=0,neq_tmp=0;i<nxsipc;i++){
            if(XCBL[i]==XCBU[i]){
                /* equality found */
                if(neq_tmp==(int)nn[0])break;
                neq_tmp++;
            }
        }
        if(i==nxsipc){
            mxErrMsgTxt("Invalid constraint number\n");
        } else {
            sip_conxgrd(i, X, CEQ, NULL);
        }
        break;
    case 8:
        /* sip_conxineqgrd */
        if(nlhs!=1 || nrhs!=3 || !mxIsDouble(prhs[1]) || !mxIsDouble(prhs[2])){
            usage();
            return;
        }
        nn=mxGetPr(prhs[1]);
        X=mxGetPr(prhs[2]);

        mx=mxGetM(prhs[2]);
        nx=mxGetN(prhs[2]);
        if((mx!=1 || nx!=nxsip) && (mx!=nxsip || nx!=1)) {
            mexPrintf("%s must be a vector of %d elements\n", mxGetName(prhs[2]), nxsip);
            usage();
            return;
        }

        C=mxGetPr(plhs[0]=mxCreateDoubleMatrix(nxsip, 1, mxREAL));
        
        for(i=0,nineq_tmp=0;i<nxsipc;i++){
            if(XCBL[i]!=XCBU[i]){
                /* inequality found */
                if(nineq_tmp==(int)nn[0])break;
                nineq_tmp++;
            }
        }
        if(i==nxsipc){
            mxErrMsgTxt("Invalid constraint number\n");
        } else {
            sip_conxgrd(i, X, C, NULL);
        }
        break;
    case 9:
        /* sip_conxeqhes */
        if(nlhs!=1 || nrhs!=2 || !mxIsDouble(prhs[1])){
            usage();
            return;
        }
        nn=mxGetPr(prhs[1]);

        CEQ=mxGetPr(plhs[0]=mxCreateDoubleMatrix(nxsip, nxsip, mxREAL));
        
        for(i=0,neq_tmp=0;i<nxsipc;i++){
            if(XCBL[i]==XCBU[i]){
                /* equality found */
                if(neq_tmp==(int)nn[0])break;
                neq_tmp++;
            }
        }
        if(i==nxsipc){
            mxErrMsgTxt("Invalid constraint number\n");
        } else {
            sip_conxhes(i, CEQ, NULL);
        }
        break;
    case 10:
        /* sip_conxineqhes */
        if(nlhs!=1 || nrhs!=2 || !mxIsDouble(prhs[1])){
            usage();
            return;
        }
        nn=mxGetPr(prhs[1]);

        C=mxGetPr(plhs[0]=mxCreateDoubleMatrix(nxsip, nxsip, mxREAL));
        
        for(i=0,nineq_tmp=0;i<nxsipc;i++){
            if(XCBL[i]!=XCBU[i]){
                /* inequality found */
                if(nineq_tmp==(int)nn[0])break;
                nineq_tmp++;
            }
        }
        if(i==nxsipc){
            mxErrMsgTxt("Invalid constraint number\n");
        } else {
            sip_conxhes(i, C, NULL);
        }
        break;
    case 11:
        /* sip_contval */
        if(nlhs!=1 || nrhs!=4 || !mxIsDouble(prhs[1]) || !mxIsDouble(prhs[2]) || !mxIsDouble(prhs[3])){
            usage();
            return;
        }
        nn=mxGetPr(prhs[1]);
        X=mxGetPr(prhs[2]);
        T=mxGetPr(prhs[3]);        

        mx=mxGetM(prhs[2]);
        nx=mxGetN(prhs[2]);
        mt=mxGetM(prhs[3]);
        nt=mxGetN(prhs[3]);
        if((mx!=1 || nx!=nxsip) && (mx!=nxsip || nx!=1)) {
            mexPrintf("%s must be a vector of %d elements\n", mxGetName(prhs[2]), nxsip);
            usage();
            return;
        }
        if((mt!=1 || nt!=ntsip) && (mt!=ntsip || nt!=1)) {
            mexPrintf("%s must be a vector of %d elements\n", mxGetName(prhs[3]), ntsip);
            usage();
            return;
        }

        RT=mxGetPr(plhs[0]=mxCreateDoubleMatrix(1, 1, mxREAL));

        if(nn[0]>=ntsipc){
            mxErrMsgTxt("Invalid constraint number\n");
        } else {        
            RT[0]=sip_contval((int)nn[0], X, T, NULL)-TCBU[(int)nn[0]];
        }
        break;
    case 12:
        /* sip_contgrd */
        if(nlhs!=2 || nrhs!=4 || !mxIsDouble(prhs[1]) || !mxIsDouble(prhs[2]) || !mxIsDouble(prhs[3])){
            usage();
            return;
        }
        nn=mxGetPr(prhs[1]);
        X=mxGetPr(prhs[2]);
        T=mxGetPr(prhs[3]);        

        mx=mxGetM(prhs[2]);
        nx=mxGetN(prhs[2]);
        mt=mxGetM(prhs[3]);
        nt=mxGetN(prhs[3]);
        if((mx!=1 || nx!=nxsip) && (mx!=nxsip || nx!=1)) {
            mexPrintf("%s must be a vector of %d elements\n", mxGetName(prhs[2]), nxsip);
            usage();
            return;
        }
        if((mt!=1 || nt!=ntsip) && (mt!=ntsip || nt!=1)) {
            mexPrintf("%s must be a vector of %d elements\n", mxGetName(prhs[3]), ntsip);
            usage();
            return;
        }

        RX=mxGetPr(plhs[0]=mxCreateDoubleMatrix(nxsip, 1, mxREAL));
        RT=mxGetPr(plhs[1]=mxCreateDoubleMatrix(ntsip, 1, mxREAL));        

        if(nn[0]>=ntsipc){
            mxErrMsgTxt("Invalid constraint number\n");
        } else {        
            sip_contgrd((int)nn[0], X, T, RX, RT, NULL);
        }
        break;
    case 13:
        /* sip_conthes */
        if(nlhs!=3 || nrhs!=2 || !mxIsDouble(prhs[1])){
            usage();
            return;
        }
        nn=mxGetPr(prhs[1]);        

        RX=mxGetPr(plhs[0]=mxCreateDoubleMatrix(nxsip, nxsip, mxREAL));
        RT=mxGetPr(plhs[1]=mxCreateDoubleMatrix(ntsip, ntsip, mxREAL));
        RXT=mxGetPr(plhs[2]=mxCreateDoubleMatrix(nxsip, ntsip, mxREAL));
        
        if(nn[0]>=ntsipc){
            mxErrMsgTxt("Invalid constraint number\n");
        } else {
            sip_conthes((int)nn[0], RX, RT, RXT, NULL);
        }
        break;
    case 14:
        /* sip_jacval */
        if(nlhs!=4 || nrhs!=3 || !mxIsDouble(prhs[1]) || !mxIsDouble(prhs[2])){
            usage();
            return;
        }
        X=mxGetPr(prhs[1]);
        T=mxGetPr(prhs[2]);

        mx=mxGetM(prhs[1]);
        nx=mxGetN(prhs[1]);
        mt=mxGetM(prhs[2]);
        nt=mxGetN(prhs[2]);
        if((mx!=1 || nx!=nxsip) && (mx!=nxsip || nx!=1)) {
            mexPrintf("%s must be a vector of %d elements\n", mxGetName(prhs[1]), nxsip);
            usage();
            return;
        }
        if((mt!=1 || nt!=ntsip) && (mt!=ntsip || nt!=1)) {
            mexPrintf("%s must be a vector of %d elements\n", mxGetName(prhs[2]), ntsip);
            usage();
            return;
        }

        if(neq==-1 || nineq==-1){
            /* count how many equality constraints */
            for(i=0,neq=0;i<nxsipc;i++)
                if(XCBL[i]==XCBU[i])neq++;
            nineq=nxsipc-neq;
        }

        C=mxGetPr(plhs[0]=mxCreateDoubleMatrix(nineq, nxsip, mxREAL));
        CEQ=mxGetPr(plhs[1]=mxCreateDoubleMatrix(neq, nxsip, mxREAL));
        RT=mxGetPr(plhs[2]=mxCreateDoubleMatrix(ntsipc, ntsip, mxREAL));
        RXT=mxGetPr(plhs[3]=mxCreateDoubleMatrix(ntsipc, nxsip, mxREAL));
        RX=mymalloc(nxsip*nxsipc*sizeof(double));
        
        sip_jacval(X, T, RX, RT, RXT, NULL);
        for(i=0,neq_tmp=0,nineq_tmp=0;i<nxsipc;i++){
            if(XCBL[i]!=XCBU[i]){
                for(j=0;j<nxsip;j++)
                    C[j*nineq+nineq_tmp]=RX[j*nxsipc+i];
                nineq_tmp++;
            } else {
                for(j=0;j<nxsip;j++)
                    CEQ[j*neq+neq_tmp]=RX[j*nxsipc+i];
                neq_tmp++;
            }
        }

        free(RX);
        break;
    case 15:
        /* sip_end, write_sol for AMPL */
        if(nrhs!=3 || !mxIsChar(prhs[1]) || !mxIsDouble(prhs[2])){
            usage();
            return;
        }
        X=mxGetPr(prhs[2]);
        mxGetString(prhs[1], msgbuf, sizeof(msgbuf));
        write_sol(msgbuf, X, NULL, NULL);
        clean_house();
        break;      
    default:
        usage();
  }

  return;

}


#if defined __linux__ || defined linux
void __sigsetjmp (sigjmp_buf env, int savemask){
}
#endif
