#include <stdio.h>
#include <stdlib.h>

#if defined __linux__ || defined linux
  #include <strings.h>
#endif

#include <math.h>
#include <getstub.h>
#undef X0
#include "nsips.h"

#ifndef SIP_SOLVER_included
#define SIP_SOLVER_included


enum {
  /* Hettich version */
  DISC_METHOD,
  DISC_METHOD_HALTON,
  /* Reemtsen version */
  DISC_METHOD_REEM,
  /* Penalty Function Solver */
  PFS,
  /* Dual Penalty Function Solver */
  PFS_DUAL,
  /* Lagrangian Penalty Function Solver */
  LPFS,
  /* Powell method with Penalty Function */
  PFPOWELLS,
  /* PFS Teo */
  PFS_TEO,
  /* SQP with integral */
  SQP,
  /* NPSOL with PF */
  NPSOL_PF,
  /* SQP DUAL with LAGrangian */
  SQP_DUAL_LAG,
  /* Tseng penalty function */
  PF_TSENGS,
  /* Filter */
  FILTER,
  /* interior-point method */
  INTP
  /*  EXCHANGE_METHOD=3,
      REDUCION_METHOD=4*/
};


enum {
  /* primal ordering */
  PRIMAL,
  /* dual ordering */
  DUAL,
  /* Predictor Corrector */
  PREDCORR
};


#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE 1
#endif


#ifndef NO_DERIV
  /* no derivatives */
#define NO_DERIV 0
  /* first derivatives */
#define FIRST_DERIV 1
#endif


enum {
  /* penalty based in plus function */
  PF_GXTPLUS,
  /* penalty based in square plus function */
  PF_GXTPLUS2,
  /* penalty based in epsilon function */
  PF_GXTEPSILON,
  /* penalty based in exponencial function */
  PF_GXTEXPPLUS,
  /* penalty based in exponencial function with epsilon */
  PF_GXTEXPEPSILON,
  /* penalty based in Teo smoothing function */
  PF_GXTTEO,
  /* penalty based in exponential and Teo smoothing function */
  PF_GXTEXPTEO,
  /* penalty based in exponential and Teo smoothing function */
  PF_EXPGXTTEO,
  /* penalty based in exponential from Tseng and Teo smoothing function */
  PF_GXTTSENGTEO,
  /* penalty based in exponential from Tseng and Teo smoothing function */
  PF_TSENGGXTTEO,
  /* penalty based in Hartman paper */
  PF_GXTHART,
  /* penalty based in Grossmann */
  PF_GXTGROSS
};

enum {
  /* integral evaluated by gaussian formula */
  PF_GAUSS,
  /* integral evaluated by trapezoid formula */
  PF_TRAP,
  /* integral evaluated by adaptative trapezoid formula */
  PF_ADAPT,
  /* integral evaluated by adaptative gaussian formula */
  PF_ADAPT_GAUSS,
  /* integral evaluated by Monte Carlo pseudo-random algorithm */
  PF_MULTINT
};


enum {
  /* line search by Armijo rule */
  ARMIJO,
  /* line search by Armijo/Goldstein */
  ARGO,
  /* line search be Wolfe */
  WOLFE,
  /* line search with armijo and bakry merit function */
  ARMI_BAKRY,
  /* line search with armijo, primal and dual step */
  ARMI_DUAL,
  /* l_2 merit function */
  L2MERIT,
  /* Augmented Lagrangian merit function */
  ALMERIT
};


enum {
  /* multiplier method lag(k+1)=lag(k)-miu*g(k) */
  MULTI,
  /* multiplier updated with generalized inverse lag(k+1)=(DcDcT)-1Dc */
  GENINV
};


typedef struct SSO {
  int disc_k; /* discretization number of refines */
  double *disc_h; /* array of step in mesh */
  double disc_h_aux; /* auxiliary step in mesh */
  int ntvars; /* dimension of h array */
  int disc_h_mx; /* maximum number os halton points */
  int disc_ha_mx; /* maximum number os halton points added in each inter */
  int disc_inner; /* maximum number of inner iterations */
  double disc_dist; /* neighbor distance in first mesh*/
  double disc_eps; /* starting value for epsilon in Reemtsen version */

  double zero;            /* points precision (zero)*/
  double reseth;          /* precision for reset to hessian inverse*/
  int dampedBFGS;         /* Use damped BFGS formula */
  int scaleBFGS;          /* Use scaled BFGS formula */
  int ScaleDir;           /* Scale direction */
  int multipliersUPD;     /* type of multipliers update */
  double ScaleDir1;       /* Lower limit for scale direction */
  double ScaleDir2;       /* upper limit for scale direction */
  double fdiff;           /* finite difference displacement */

  int maxitero;           /* maximum of outer iterations */
  int maxiteri;           /* maximum of inner iterations */

  double pf_preci;        /* stopping criteria precision inner*/
  double pf_preco;        /* stopping criteria precision outer*/
  double pf_epsilon;      /* epsilon in penalty function */
  double pf_tau;          /* tau for lic */
  double startmiu;        /* starting  miu */
  double miufactor;       /* factor for miu */
  double epsfactor;       /* factor for epsilon */
  double epslimit;        /* epsilon limit */
  int pf_type;            /* penalty function type */
  int pf_int;             /* penalty function integral type */
  int n_int;              /* number of points in trapezoids formula */
  int adapt_int_n;        /* n. of starting points in adaptative formulas */
  double adapt_int_amp;   /* maximum set amplitude for adaptative formulas */
  double adapt_int_error; /* maximum error for adaptative formulas */

  int powell_nmax;        /* maximum iteration in Powell method */
  double powell_epson;    /* epson from stopping criteria */

  int derivatives;        /* derivatives level */

  int linesearch;         /* line search type */
  double armijo;          /* Armijo line search constant */
  double argo1;           /* Goldstein Armijo line search constant 1*/
  double argo2;           /* Goldstein Armijo line search constant 2*/
  double wolfe1;          /* Armijo Wolfe line search constant 1*/
  double wolfe2;          /* Armijo Wolfe line search constant 2*/

  int reset_h;            /* reset hessian inverse */
  int dual_ini_res;       /* reset dual initial to QP */

  int debug_level;        /* bit 0 print one line outer iteration 
			     bit 1 print one line per inner iteration 
			  */

  double theta;           /* for initial guess of slack varaibles */
  double delta;           /* delta for the barrier parameter */
  int updtb;              /* which barrier parameter formula */
  int method;             /* method used to solve SIP */
  double int_prec;        /* stopping criteria precision for IPM */

  int watchdog;           /* with or without watchdog */
  int watchmax;           /* watchdog limit in merit function */

  double choleps;         /* small epsilon in cholesky factorization */

  int ordering;           /* ordering in solving KKT system */
  int ncorrector;
} SSO;


typedef struct Points {
  double *coord; /* Points Coordinates */
  double norm; /* norm 2 of point */
  double *gval;
  int k_Mesh; /* in what Mesh this point appeared */
  struct Points *next; /* next Point in Mesh */
} Points;
  

typedef struct Mesh {
  double Min_Norm, Max_Norm;
  double AmpHash;
  int nHash;
  int nPoints; /* number of points in mesh */
  Points **Mesh; /* Pointers to Points structures */
} Mesh;


/* I need to make this data externally available */
typedef struct SIP_S {
  int tvars;
  int xvars;
  int nxcon;
  int ntcon;
  double *X;
  double *xbl, *xbu;
  double *tbl, *tbu;
  double *xcbl, *xcbu;
  double *tcbl, *tcbu;
  double (*sip_eval)();
} SIP_S;


#define nxsip sip_s.xvars
#define ntsip sip_s.tvars
#define nxsipc sip_s.nxcon
#define ntsipc sip_s.ntcon
#define X sip_s.X
#define XBU sip_s.xbu
#define XBL sip_s.xbl
#define TBU sip_s.tbu
#define TBL sip_s.tbl
#define XCBU sip_s.xcbu
#define XCBL sip_s.xcbl
#define TCBU sip_s.tcbu
#define TCBL sip_s.tcbl
#define eval_sip (sip_s.sip_eval)



void sip_option(void);
void sip_solver(int Lnxsip,double *LXBU,double *LXBL,int Lntsip,double *LTBU,
		double *LTBL,int Lnxsipc,double *LXCBU,double *LXCBL,
		int Lntsipc,double *LTCBU,double *LTCBL,double *LX,int guess,
		double (*Lsip_eval) ());

void check_onedim_finite(void);
void check_multidim_finite(void);
void check_no_finite_const(void);
void check_const_leq(void);
void check_finite_bound(void);

void discretization(void);
void disc_halton(void);
void fill_mesh(double *h);
void fill_R(void);
void fill_T(void);
int fill_S(void);
void print_mesh(Mesh *mset);
double norm2(int n, double *x);
void gennext(int n, double *t, double *h);
void insertpt(double *t, double norm, double *gval, Mesh *mset);
int check_insertpt(double *t, double norm, double *gval, Mesh *mset);
double solve_nlp(void);
int full_mesh(void);
void augment_T(void);
void augment_R(void);
void augment_R_RX(void);
void next_T(double *h, int k);
#ifndef mymalloc
#define mymalloc(x) mymalloc_ASL(x)
extern void *mymalloc_ASL(size_t len);
#endif
/*  void *mymalloc(int n); */
int equal(double *a, double *b, int n);
void free_mesh(Mesh *a);
void copy_mesh(Mesh *a, Mesh *b);
void add_mesh(Mesh *a, Mesh *b);
void next_halton_T(int *ht);
void fill_halton_mesh(int *ht);
void check_nlp(void);
int neighbor(double *a, double *b, int n);
void insert_neig(double *t, double norm, Mesh *mset);
void fill_n_T(double epson, double fx);
void disc_reem(void);
int check_solve(void);
void sip_iguess(double *x);
void check_T(void);
void write_result(double fx);

void penaltyfs(void);
void penaltyfs_dual(void);
void penaltylfs(void);
void penaltyfs_teo(void);
void QPs(void);
void pfpowells(void);
void npsol_penaltyfs(void);
void dual_lags(void);
void pf_tsengs(void);
void filters(void);
void intpkkts(void);

char *sip_opt_s(Option_Info *oi, keyword *kw, char *value);
char *sip_opt_i(Option_Info *oi, keyword *kw, char *value);
char *sip_opt_d(Option_Info *oi, keyword *kw, char *value);



#define maximum(a, b) (a>b ? a:b)
#define minimum(a, b) (a>b ? b:a)
#define HASH_SIZE 100

#endif
