/*******************************************************************

sip.c = Semi-Infinite Programming

This file adds extra functions to ampl solver interface
It provides Semi-Infinite Programming for ampl

Ismael Vaz -- aivaz@dps.uminho.pt

Provided unctions:
  sip_init     sip_extractx  sip_extractt  sip_joinxt   sip_objval  sip_objgrd
  sip_objhes   sip_conval    sip_conxval   sip_contval  sip_jacval  sip_conxgrd
  sip_contgrd  sip_conxhes   sip_conthes   sip_free

********************************************************************/

#include "asl.h"
#include "sip.h"

/* for debug info*/
int sip_debug=0;

extern ASL *asl;

/*********************************
Count t and x vars. If t vars are zero then this is not a SIP.
*********************************/
int sip_init(ASL *asl)
{
  int i, lastt, lastx, t, x;
  char *vname, *cname;

  ntsip=0;
  nxsip=0;

  /* count how many t and x vars are there */

  for(i=0;i<n_var;i++){
    vname=var_name(i);
    if(!strncmp(vname,"t", 1)){
      ntsip++;
    } else {
      nxsip++;
    }
    if(sip_debug)
	  printf("Var %d name is: %s\n", i+1, vname);
  }

  /* allocate diference array */
  tsipd=(int *)Malloc(ntsip*sizeof(int));
  xsipd=(int *)Malloc(nxsip*sizeof(int));

  /* allocate bounds arrays */
  XBU=REAL(nxsip);
  XBL=REAL(nxsip);
  TBU=REAL(ntsip);
  TBL=REAL(ntsip);

  /* fill difference array */
  for(i=0, lastt=0, lastx=0, t=0, x=0;i<n_var;i++){
    vname=var_name(i);
    if(!strncmp(vname,"t", 1)){
      tsipd[t]=i-lastt;
      lastt=i;
      /* lower and upper bounds */
      TBU[t]=LUv[2*i+1];
      TBL[t++]=LUv[2*i];
    } else {
      xsipd[x]=i-lastx;
      lastx=i;
      /* lower and upper bounds */
      XBU[x]=LUv[2*i+1];
      XBL[x++]=LUv[2*i];
    }
  }


  /* count how many t constraints are there */
  ntsipc=0;
  nxsipc=0;

  for(i=0;i<n_con;i++){
    cname=con_name(i);
    if(!strncmp(cname,"t", 1)){
      ntsipc++;
    } else {
      nxsipc++;
    }
    if(sip_debug)
      printf("Const %d name is: %s\n", i+1, cname);
  }

  /* allocate diference array */
  tsipcd=(int *)Malloc(ntsipc*sizeof(int));
  xsipcd=(int *)Malloc(nxsipc*sizeof(int));

  /* allocate bounds arrays */
  XCBU=REAL(nxsipc);
  XCBL=REAL(nxsipc);
  TCBU=REAL(ntsipc);
  TCBL=REAL(ntsipc);

  /* fill difference array */
  for(i=0, lastt=0, lastx=0, t=0, x=0;i<n_con;i++){
    cname=con_name(i);
    if(!strncmp(cname,"t", 1)){
      tsipcd[t]=i-lastt;
      lastt=i;
      TCBU[t]=LUrhs[2*i+1];
      TCBL[t++]=LUrhs[2*i];
    } else {
      xsipcd[x]=i-lastx;
      lastx=i;
      XCBU[x]=LUrhs[2*i+1];
      XCBL[x++]=LUrhs[2*i];
    }
  }

  dense_j();

  
  if(sip_debug){
    printf("Variables t %d x %d\n", ntsip, nxsip);
    printf("\nVariables t difference array\n");
    for(i=0;i<ntsip;i++)
      printf("%d ",tsipd[i]);
    
    printf("\nVariables x difference array\n");
    for(i=0;i<nxsip;i++)
      printf("%d ",xsipd[i]);
    printf("\n");

    printf("Constraints t %d x %d\n", ntsipc, nxsipc);
    printf("\nConstraints t difference array\n");
    for(i=0;i<ntsipc;i++)
      printf("%d ",tsipcd[i]);
    
    printf("\nConstraints x difference array\n");
    for(i=0;i<nxsipc;i++)
      printf("%d ",xsipcd[i]);
    printf("\n");

    printf("\nVariables bounds t\n");
    for(i=0;i<ntsip;i++)
      printf("%lf -- %d -- %lf\n",TBL[i],i+1,TBU[i]);

    printf("\nVariables bounds x\n");
    for(i=0;i<nxsip;i++)
      printf("%lf -- %d -- %lf\n",XBL[i],i+1,XBU[i]);

    printf("\nConstraints bounds t\n");
    for(i=0;i<ntsipc;i++)
      printf("%lf -- %d -- %lf\n",TCBL[i],i+1,TCBU[i]);

    printf("\nConstraints bounds x\n");
    for(i=0;i<nxsipc;i++)
      printf("%lf -- %d -- %lf\n",XCBL[i],i+1,XCBU[i]);

  }

  /* I should check for extra stuff (non presence of t vars in objective, etc..
     but for the moment it's all */

  if(!ntsip || !ntsipc)return SIP_invalid;
  return(SIP_OK);
}

/************************
copy a to b. a is a n_vars array, b is a nxsip array
************************/
void sip_extractx(real *a, real *b)
{
  int i;
  real *tmp;

  for(i=0, tmp=a+xsipd[0];i<nxsip;i++, tmp+=xsipd[i])
    b[i]=*tmp;
}

/************************
copy a to b. a is a n_vars array, b is a nxsip array
************************/
void sip_extractt(real *a, real *b)
{
  int i;
  real *tmp;

  for(i=0, tmp=a+tsipd[0];i<ntsip;i++, tmp+=tsipd[i])
    b[i]=*tmp;
}

/************************
copy a+b to c. a is a nxsip array, b is a ntsip arrya and c is a n_vars array
x+t=x0
************************/
void sip_joinxt(real *a, real *b, real *c)
{
  int i;
  real *tmp;

  for(i=0, tmp=c+tsipd[0];i<ntsip;i++, tmp+=tsipd[i])
    *tmp=b[i];
  for(i=0, tmp=c+xsipd[0];i<nxsip;i++, tmp+=xsipd[i])
    *tmp=a[i];
}


/****************
objective function for SIP
****************/
real sip_objval(int nobj, real *X, fint *nerror)
{
  real *Y, *T;
  real fval;

  Y=REAL(n_var);
  T=REAL(ntsip);
  memset(T,0,ntsip*sizeof(real));

  /* join X+T in Y */
  sip_joinxt(X, T, Y);

  fval=objval(nobj, Y, nerror);
  free(Y);
  return fval;
}



/****************
objective gradient for SIP
****************/
void sip_objgrd(int nobj, real *X, real *G, fint *nerror)
{

  real *Y, *T, *GY;

  Y=REAL(n_var);
  GY=REAL(n_var);
  T=REAL(ntsip);
  memset(T,0,ntsip*sizeof(real));

  /* join X+T in Y */
  sip_joinxt(X, T, Y);

  objgrd(nobj, Y, GY, nerror);

  sip_extractx(GY, G);

  free(T);
  free(Y);
  free(GY);
  return;
}


/***********************
Objective Hessian for SIP
***********************/
void sip_objhes(int nobj, real *H, fint *nerror)
{
  real *HY, *tmp1, *tmp2;
  int i,j;

  /* matrix for full hessian */
  HY=REAL(n_var*n_var);

  fullhes(HY, n_var, nobj, NULL, NULL);


  for(i=0,tmp1=HY;i<nxsip;i++){
    tmp1+=xsipd[i]*n_var;
    tmp2=tmp1+xsipd[0];
    for(j=0;j<nxsip;j++,tmp2+=xsipd[j])
      H[j+i*nxsip]=*tmp2;
  }

  free(HY);
  return;
}


/**********************
Evaluate all constraints
**********************/
void sip_conval(real *X, real *T, real *RX, real *RT, fint *nerror)
{
  real *Y, *RY, *tmp;
  int i;

  Y=REAL(n_var);
  RY=REAL(n_con);

  /* joint X and T in Y */
  sip_joinxt(X, T, Y);

  conval(Y, RY, nerror);

  /* split RY vector in RX and RT */
  for(i=0,tmp=RY+xsipcd[0];i<nxsipc;i++,tmp+=xsipcd[i])
    RX[i]=*tmp;
  for(i=0,tmp=RY+tsipcd[0];i<ntsipc;i++,tmp+=tsipcd[i])
    RT[i]=*tmp;

  free(Y);
  free(RY);
}

/**********************
Evaluate finite (x) constraints
**********************/
real sip_conxval(int ncon, real *X, fint *nerror)
{
  real *Y, *T, rval;
  int i, newc;

  Y=REAL(n_var);
  T=REAL(ntsip);
  memset(T, 0, ntsip*sizeof(real));

  /* joint X and T in Y */
  sip_joinxt(X, T, Y);

  /* calculate new constraint number */
  for(i=0,newc=xsipcd[0];i<ncon;i++,newc+=xsipcd[i]);

  rval=conival(newc, Y, nerror);

  free(Y);
  free(T);
  return rval;
}


/**********************
Evaluate infinite (t) constraints
**********************/
real sip_contval(int ncon, real *X, real *T, fint *nerror)
{
  real *Y, rval;
  int i, newc;

  Y=REAL(n_var);

  /* joint X and T in Y */
  sip_joinxt(X, T, Y);

  /* calculate new constraint number */
  for(i=0,newc=tsipcd[0];i<ncon;i++,newc+=tsipcd[i]);

  rval=conival(newc, Y, nerror);

  free(Y);
  return rval;
}


/**********************
Evaluate jacobian of all variables
**********************/
void sip_jacval(real *X, real* T, real *JX, real *JT, real *JXT, fint *nerror)
{
  real *Y, *J, *tmp1, *tmp2;
  int i,j;

  Y=REAL(n_var);
  J=REAL(n_con*n_var);

  sip_joinxt(X, T, Y);

  jacval(Y, J, nerror);

  for(i=0,tmp1=J;i<nxsip;i++){
    tmp1+=xsipd[i]*n_con;
    tmp2=tmp1+xsipcd[0];
    for(j=0; j<nxsipc;j++, tmp2+=xsipcd[j])
      JX[j+i*nxsipc]=*tmp2;
  }

  for(i=0,tmp1=J;i<ntsip;i++){
    tmp1+=tsipd[i]*n_con;
    tmp2=tmp1+tsipcd[0];
    for(j=0; j<ntsipc;j++, tmp2+=tsipcd[j])
      JT[j+i*ntsipc]=*tmp2;
  }
  
  for(i=0,tmp1=J;i<nxsip;i++){
    tmp1+=xsipd[i]*n_con;
    tmp2=tmp1+tsipcd[0];
    for(j=0; j<ntsipc;j++, tmp2+=tsipcd[j])
      JXT[j+i*ntsipc]=*tmp2;
  }

  free(Y);
  free(J);
  return;
}


/**********************
Evaluate gradient of finite (x) constraint
**********************/
void sip_conxgrd(int ncon, real *X, real *G, fint *nerror)
{
  real *Y, *T, *GY, *tmp;
  int i, newc;

  Y=REAL(n_var);
  GY=REAL(n_var);
  T=REAL(ntsip);
  memset(T, 0, ntsip*sizeof(real));

  /* joint X and T in Y */
  sip_joinxt(X, T, Y);

  /* calculate new constraint number */
  for(i=0,newc=xsipcd[0];i<ncon;i++,newc+=xsipcd[i]);

  congrd(newc, Y, GY, nerror);

  /* copy x part from gradient */
  for(i=0, tmp=GY+xsipd[0];i<nxsip;i++,tmp+=xsipd[i])
    G[i]=*tmp;

  free(Y);
  free(GY);
  free(T);
  return;
}


/**********************
Evaluate gradient of infinite (t) constraint
**********************/
void sip_contgrd(int ncon, real *X, real *T, real *GX, real *GT, fint *nerror)
{
  real *Y, *GY, *tmp;
  int i, newc;

  Y=REAL(n_var);
  GY=REAL(n_var);

  /* joint X and T in Y */
  sip_joinxt(X, T, Y);

  /* calculate new constraint number */
  for(i=0,newc=tsipcd[0];i<ncon;i++,newc+=tsipcd[i]);

  congrd(newc, Y, GY, nerror);

  /* copy x part from gradient */
  for(i=0, tmp=GY+xsipd[0];i<nxsip;i++,tmp+=xsipd[i])
    GX[i]=*tmp;
  /* copy t part from gradient */
  for(i=0, tmp=GY+tsipd[0];i<ntsip;i++,tmp+=tsipd[i])
    GT[i]=*tmp;

  free(Y);
  free(GY);
  return;
}


/**********************
Evaluate hessian of finite (x) constraint
**********************/
void sip_conxhes(int ncon, real *H, fint *nerror)
{
  int newc, i, j;
  real *HY, *Y, *tmp1, *tmp2;

  /* matrix for full hessian */
  HY=REAL(n_var*n_var);
  Y=REAL(n_con);
  memset(Y, 0, n_con*sizeof(real));

  /* calculate new constraint number */
  for(i=0,newc=xsipcd[0];i<ncon;i++,newc+=xsipcd[i]);
  /* just constraint newc */
  if(newc<0 || newc>=n_con)newc=xsipcd[0];
  Y[newc]=1;

  fullhes(HY, n_var, -1, NULL, Y);

  for(i=0,tmp1=HY;i<nxsip;i++){
    tmp1+=xsipd[i]*n_var;
    tmp2=tmp1+xsipd[0];
    for(j=0;j<nxsip;j++,tmp2+=xsipd[j])
      H[j+i*nxsip]=*tmp2;
  }
  free(HY);
  free(Y);
  return;
}



/**********************
Evaluate hessian of infinite (t) constraint
**********************/
void sip_conthes(int ncon, real *HX, real *HT, real *HXT, fint *nerror)
{
  int newc, i, j;
  real *HY, *Y, *tmp1, *tmp2;

  /* matrix for full hessian */
  HY=REAL(n_var*n_var);
  Y=REAL(n_con);
  memset(Y, 0, n_con*sizeof(real));

  /* calculate new constraint number */
  for(i=0,newc=tsipcd[0];i<ncon;i++,newc+=tsipcd[i]);

  /* just constraint newc */
  if(newc<0 || newc>=n_con)newc=tsipcd[0];
  Y[newc]=1;

  fullhes(HY, n_var, -1, NULL, Y);

  for(i=0,tmp1=HY;i<nxsip;i++){
    tmp1+=xsipd[i]*n_var;
    tmp2=tmp1+xsipd[0];
    for(j=0;j<nxsip;j++,tmp2+=xsipd[j])
      HX[j+i*nxsip]=*tmp2;
    tmp2=tmp1+tsipd[0];
    for(j=0;j<ntsip;j++,tmp2+=tsipd[j])
      HXT[j+i*ntsip]=*tmp2;
  }

  for(i=0,tmp1=HY;i<ntsip;i++){
    tmp1+=tsipd[i]*n_var;
    tmp2=tmp1+tsipd[0];
    for(j=0;j<ntsip;j++,tmp2+=tsipd[j])
      HT[j+i*ntsip]=*tmp2;
  }

  free(HY);
  free(Y);
  return;

}


/*********************************
Free sip struct
*********************************/
void sip_free()
{
  if(tsipd)
    free(tsipd);
  if(xsipd)
    free(xsipd);
  if(XBU)
    free(XBU);
  if(XBL)
    free(XBL);
  if(TBU)
    free(TBU);
  if(TBL)
    free(TBL);

  if(tsipcd)
    free(tsipcd);
  if(xsipcd)
    free(xsipcd);
  if(XCBU)
    free(XCBU);
  if(XCBL)
    free(XCBL);
  if(TCBU)
    free(TCBU);
  if(TCBL)
    free(TCBL);
}
