###############################################################################
# Objective: Linear
# Constraints: General
###############################################################################
#
#   As described by A. De Luca, L. Lanari and G. Oriolo
#	in "A sensitivity approach to optimal spline robot trajectories"
#    Automatica, Vol. 27, No. 3, pp. 535-539, 1991
#
#   Two-link SCARA-type robot. Velocity and torque constraints
#
#   Example 1
#
#    Coded by A. Ismael F. Vaz 27/03/2003  aivaz@dps.uminho.pt
#    University of Minho, Portugal
###############################################################################
function cspline;

# number of joints
param nj;

# Initial and final velocity and acceleration for each joint
param vi{1..nj};
param vf{1..nj};
param ai{1..nj};
param af{1..nj};

# Bound limits, velocity and torque
param limits{1..2,1..nj};

# Number of cubic sub-splines
# Number of finite variables
param nx;

# Number of spline parameters
# Number of infinite variables
param nt;

# Via points in joint space in rad
param knots{1..nx-1,1..nj};

# initial guess
param xinit{1..nx};
param tinit{1..nt};

# x variables
var x {i in 1..nx} := xinit[i];
# t variable
var t {j in 1..nt} := tinit[j];


# Robot parameters
param l1;
param l2;
param d2;
param m2;
param mp;
param J1;
param J2;
param Jp;

# Cubic splines for dynamic equation
var q1   = cspline(t[1]*(sum {i in 1..nx} (x[i])), 0, nx, {i in 1..nx} x[i],
	 {i in 1..nx-1} knots[i,1], vi[1], vf[1], ai[1], af[1]);
var q2   = cspline(t[1]*(sum {i in 1..nx} (x[i])), 0, nx, {i in 1..nx} x[i],
	 {i in 1..nx-1} knots[i,2], vi[2], vf[2], ai[2], af[2]);

var q1d1 = cspline(t[1]*(sum {i in 1..nx} (x[i])), 1, nx, {i in 1..nx} x[i],
	 {i in 1..nx-1} knots[i,1], vi[1], vf[1], ai[1], af[1]);
var q2d1 = cspline(t[1]*(sum {i in 1..nx} (x[i])), 1, nx, {i in 1..nx} x[i],
	 {i in 1..nx-1} knots[i,2], vi[2], vf[2], ai[2], af[2]);

var q1d2 = cspline(t[1]*(sum {i in 1..nx} (x[i])), 2, nx, {i in 1..nx} x[i],
	 {i in 1..nx-1} knots[i,1], vi[1], vf[1], ai[1], af[1]);
var q2d2 = cspline(t[1]*(sum {i in 1..nx} (x[i])), 2, nx, {i in 1..nx} x[i],
	 {i in 1..nx-1} knots[i,2], vi[2], vf[2], ai[2], af[2]);

# Dynamic equation for torque
param H1 := J1+J2+Jp+m2*l1*l1+mp*(l1*l1+l2*l2);
param H2 := m2*l1*d2+mp*l1*l2;
param H3 := J2+Jp+mp*l2*l2;

var u1 = (H1+2*H2*cos(q2))*q1d2+(H3+H2*cos(q2))*q2d2
	  -H2*(2*q1d1*q2d1+q2d1*q2d1)*sin(q2);
var u2 = (H3+H2*cos(q2))*q1d2+H3*q2d2+H2*q1d2*q1d1*sin(q2);


minimize fx:
	sum {i in 1..nx} (x[i]);


# These constraint can be used to display the cubic splines under MATLAB

#subject to tdrawpos {j in 1..nj}:
#	 cspline(t[1]*(sum {i in 1..nx} (x[i])), 0, nx,
#	 {i in 1..nx} x[i],
#	 {i in 1..nx-1} knots[i,j], vi[j], vf[j], ai[j], af[j]) <= 0;
#subject to tdrawvel {j in 1..nj}:
#	 cspline(t[1]*(sum {i in 1..nx} (x[i])), 1, nx,
#	 {i in 1..nx} x[i],
#	 {i in 1..nx-1} knots[i,j], vi[j], vf[j], ai[j], af[j]) <= 0;
#subject to tdrawacc {j in 1..nj}:
#	 cspline(t[1]*(sum {i in 1..nx} (x[i])), 2, nx,
#	 {i in 1..nx} x[i],
#	 {i in 1..nx-1} knots[i,j], vi[j], vf[j], ai[j], af[j]) <= 0;
#subject to tdrawjerk {j in 1..nj}:
#	 cspline(t[1]*(sum {i in 1..nx} (x[i])), 3, nx,
#	 {i in 1..nx} x[i],
#	 {i in 1..nx-1} knots[i,j], vi[j], vf[j], ai[j], af[j]) <= 0;
#subject to tdrawtorque1:
#	u1 <=0;
#subject to tdrawtorque2:
#	u2 <=0;


subject to tvel1:
	-limits[1,1] <= q1d1 <= limits[1,1];

subject to tvel2:
	-limits[1,2] <= q2d1 <= limits[1,2];

subject to torque1:
	-limits[2,1] <= u1 <= limits[2,1];

subject to torque2:
	-limits[2,2] <= u2 <= limits[2,2];


subject to bounds {i in 1..nx}:
	x[i] >= 0.01;
subject to tbounds {i in 1..nt}:
	0 <= t[i] <= 1;

data;

param nx := 7;
param nt := 1;

# Initial and final velocity and acceleration
param vi :=
	1 0
	2 0;
param vf :=
	1 0
	2 0;
param ai := 
	1  13.880794
	2 -11.067942;
param af := 
	1 -0.415203
	2 -4.186542;

# Robot parameters
param l1 := 0.5; 
param l2 := 0.5;
param d2 := 0.25;
param m2 := 1;
param mp := 0;
param J1 := 0.084;
param J2 := 0.084;
param Jp := 0;


# initial guess
param xinit:=
	1 1
	2 1	# extra space for extra freedom
	3 0.5
	4 0.5
	5 0.5
	6 0.5	# extra space for extra freedom
	7 0.5;


# De Luca solution as initial guess
#param xinit :=
#	1	0.01
#	2	0.37
#	3	0.25
#	4	0.34
#	5	0.43
#	6	1.07
#	7	0.01;
	


# NSIPS solution, disc_hett method with default options
#param xinit:=
#	1  0.01
#	2  0.348255
#	3  0.260631
#	4  0.361528
#	5  0.351404
#	6  0.01
#	7  1.06135;



param tinit:=
	1 0;

param nj  := 2;

# Via points in rad
param knots :
	1	2 :=
1	0	0
2	0.5	-0.5
3	0.75	-1
4	1	-1.5
5	1.25	-1
6	1.5	0.5;


# Bound limits
param limits:
	1	2 :=
1	2	2  # velocity
2	7	2;  # torque


# solve problem
# don't forget to write .col and .row files
option nsips_auxfiles rc;
option solver nsips;
solve;

##################
# Solution found #
##################
printf "Solution found\n";
display x;
display fx;

################################################
# solution of problem presented by the authors #
################################################
printf "Original Solution\n";
printf "h= %lf, %lf, %lf, %lf, %lf\n", 0.37, 0.25, 0.34, 0.43, 1.07;
printf "Total time= %lf\n", 2.46;
