# Same as elke4, but with constraints in standard form (<=)
###############################################################################
# Objective: Linear
# Constraints: Generic
###############################################################################
#  Robotics path planning problem
#   As described by Elke Haaren-Retagne, PhD Thesis,
#                 submitted to Trier University, "A Semi-Infinite Programming
#                                   "Algorithm for Robot Trajectory Planning",
#                 1992
#   Originaly described Marin
#
#   Example 2 (model 2)
#    Coded by A. Ismael F. Vaz 27/02/2000  aivaz@dps.uminho.pt
#    University of Minho, Portugal
###############################################################################
function bspline;
function dbspline;


param k;            # spline degree
param n;            # number of blend splines
# The number of knots can be changed, do not forget to change the Knots array
param nk;           # number of knots
param Knots{1..nk}; # knots
param x0{1..n};     # starting guess
param t0;
param epson;
param pi := atan(1)*4;
param grav := 9.81;

# robot constants
param J{1..3};
param nj{1..3};
param D{1..3};
param C{1..3};

# robot parameters
param rm{1..2}; # robot masses
param rd{1..2}; # robot links lengths


var x{i in 1..n} := x0[i];  # x variables (coeficients in the spline)
var t := t0;                # spline parameter

# describe trajectory
var zeta = -2*t^3+3*t^2;
var zetad1 = -6*t^2+6*t; # first derivative
var zetad2 = -12*t+6;    # second derivative
param zetad3 := -12;        # third derivative

var teta1 = (4*pi/3)*(zeta-0.5)^2;
var teta1d1 = (8*pi/3)*(zeta-0.5)*zetad1;
var teta1d2 = (8*pi/3)*(zetad1^2+(zeta-0.5)*zetad2);
var teta1d3 = (8*pi/3)*(3*zetad1*zetad2+(zeta-0.5)*zetad3);

var teta2 = -(pi/2)+(pi/4)*sin(2*pi*zeta);
var teta2d1 = (pi^2/2)*cos(2*pi*zeta)*zetad1;
var teta2d2 = (pi^2/2)*(-2*pi*sin(2*pi*zeta)*zetad1^2+cos(2*pi*zeta)*zetad2);
var teta2d3 = (pi^2/2)*(-4*pi^2*cos(2*pi*zeta)*zetad1^3-
		6*pi*sin(2*pi*zeta)*zetad1*zetad2+cos(2*pi*zeta)*zetad3);

var teta3 = (pi/2)*zeta;
var teta3d1 = (pi/2)*zetad1;
var teta3d2 = (pi/2)*zetad2;
param teta3d3 := (pi/2)*zetad3;

var g =	   bspline(t,    k, n, {i in 1..n} x[i], {i in 1..nk} Knots[i]);
var gd1 = dbspline(t, 1, k, n, {i in 1..n} x[i], {i in 1..nk} Knots[i]);


# auxiliary equation
var c1 = cos(teta1);
var c2 = cos(teta2);
var c12= cos(teta1+teta2);
var s1 = sin(teta1);
var s2 = sin(teta2);
var s12= sin(teta1+teta2);

var K1    = 0.5*(rm[1]+rm[2])*rd[1]^2+0.5*(rm[2]*rd[2]^2)+rm[2]*rd[1]*rd[2]*c2;
param K2 := 0.5*rm[2]*rd[2]^2;
var K3    = 0.5*(rm[1]+rm[2])*rd[1]^2*c1^2+0.5*rm[2]*rd[2]^2*c12^2+
	    rm[2]*rd[1]*rd[2]*c1*c12;
var K4    = rm[2]*rd[2]^2+rm[2]*rd[1]*rd[2]*c2;

var K1t2  = -rd[1]*rd[2]*rm[2]*s2;
var K3t1  = -(rm[1]*rm[2])*rd[1]^2*c1*s1-rm[2]*rd[1]*rd[2]*(s1*c12+c1*s12)-
	    rd[2]^2*rm[2]^2*c12*s12;
var K3t2  = -rd[1]*rd[2]*rm[2]*s2;
var K4t2  = K3t2;
var Pt1   = (rm[1]+rm[2])*grav*rd[1]*c1+rm[2]*grav*rd[2]*c12;
var Pt2   = rm[2]*grav*rd[2]*c12;

var u1    = 2*K1*teta1d2+K4*teta2d2+K4t2*teta2d1^2-K3t1*teta3d1^2+
	    2*K1t2*teta1d1*teta2d1;
var u2    = K4*teta1d2+2*K2*teta2d2-K1t2*teta1d1^2-K3t2*teta3d1^2;
var u3    = 2*K3*teta3d2+2*K3t1*teta1d1*teta3d1+2*K3t2*teta2d1*teta3d1;

var v1    = 2*K1*teta1d1+K4*teta2d1;
var v2    = K4*teta1d1+2*K2*teta2d1;
var v3    = 2*K3*teta3d1;

var w1    = Pt1;
var w2    = Pt2;
param w3 := 0;


minimize obj:
	(1/k)*(sum {i in 1..n} (x[i]*(Knots[i+k]-Knots[i])));

subject to bound:
	0 <= t <= 0.999999;

subject to tcons1a:
	(J[1]*nj[1]*((teta1d2-teta1d1*(gd1/g))/g^2)+
	       D[1]*nj[1]*(teta1d1/g)+
	       (((u1-v1*(gd1/g))/g^2)+w1)/nj[1])/C[1] <=1;
subject to tcons1b:
	-(J[1]*nj[1]*((teta1d2-teta1d1*(gd1/g))/g^2)+
	       D[1]*nj[1]*(teta1d1/g)+
	       (((u1-v1*(gd1/g))/g^2)+w1)/nj[1])/C[1] <=1;
subject to tcons2a:
	(J[2]*nj[2]*((teta2d2-teta2d1*(gd1/g))/g^2)+
	       D[2]*nj[2]*(teta2d1/g)+
	       (((u2-v2*(gd1/g))/g^2)+w2)/nj[2])/C[2] <=1;
subject to tcons2b:
	-(J[2]*nj[2]*((teta2d2-teta2d1*(gd1/g))/g^2)+
	       D[2]*nj[2]*(teta2d1/g)+
	       (((u2-v2*(gd1/g))/g^2)+w2)/nj[2])/C[2] <=1;
subject to tcons3a:
	(J[3]*nj[3]*((teta3d2-teta3d1*(gd1/g))/g^2)+
	       D[3]*nj[3]*(teta3d1/g)+
	       (((u3-v3*(gd1/g))/g^2)+w3)/nj[3])/C[3] <=1;
subject to tcons3b:
	-(J[3]*nj[3]*((teta3d2-teta3d1*(gd1/g))/g^2)+
	       D[3]*nj[3]*(teta3d1/g)+
	       (((u3-v3*(gd1/g))/g^2)+w3)/nj[3])/C[3] <=1;

subject to tcons4:
	-g + epson <= 0;

data;


param J :=
	1 1.98E-4
	2 1.98E-4
	3 1.98E-4;

param nj:=
	1 62.5
	2 62.5
	3 62.5;

param D :=
	1 1.744E-3
	2 1.744E-3
	3 1.744E-3;

param C :=
	1 2.090
	2 0.523
	3 1.045;

param rm :=
	1 5
	2 2;

param rd :=
	1 1.5
	2 0.5;

param k  := 4;
param n  := 9;
param nk := 13;
param epson:= 1D-2;

param Knots :=
	1	0
	2	0
	3	0
	4	0
	5	0.166666
	6	0.333332
	7	0.499998
	8	0.666664
	9	0.833330
	10	1
	11	1
	12	1
	13	1;

param x0 :=
	1	3.90451
	2	3.90451
	3	3.90451
	4	3.90451
	5	3.90451
	6	3.90451
	7	3.90451
	8	3.90451
	9	3.90451;


param t0 := 0;


# solve problem
# don't forget to write .col and .row files
option nsips_auxfiles rc;
option solver nsips;
solve;

##################
# Solution found #
##################
printf "Solution found\n";
display x;
display obj;

################################################
# solution of problem presented by the authors #
################################################
printf "Original Solution\n";
printf "obj=%lf\n", 2.48325;

