###############################################################################
# Objective: Linear
# Constraints: Polynomial
###############################################################################
#
#   As described by C.-S. Lin, P.-R. Chang and J.Y.S. Luh
#	in "Formulation and Optimization of Cubic Polynomial Joint
#		Trajectories for Industrial Robots",
#    IEEE Trans. Automatic Control, Vol. AC-28, No. 12, December 1983
#
#   Unimate PUMA 560 robot with 6 revolute joints. Simple bound constraints
#      on velocity, acceleration and jerk
#
#    Coded by A. Ismael F. Vaz 15/03/2003  aivaz@dps.uminho.pt
#    University of Minho, Portugal
###############################################################################
function cspline;

# number of joints
param nj;

# Initial and final velocity and acceleration
param vi;
param vf;
param ai;
param af;

# Initial time
param it;

# Bound limits, velocity, acceleration and jerk
param limits{1..3,1..nj};

# Number of cubic sub-splines
# Number of finite variables
param nx;

# Number of spline parameters
# Number of infinite variables
param nt;

# Via points in joint space in degrees
param knots{1..nx-1,1..nj};


# initial guess
param xinit{1..nx};
param tinit{1..nt};

# x variables
var x {i in 1..nx} := xinit[i];
# t variable
var t {j in 1..nt} := tinit[j];

# linear change of variable
#var tt = t[1]*(sum {i in 1..nx} (x[i]));

minimize fx:
	sum {i in 1..nx} (x[i]);


# These constraint can be used to display the cubic splines under MATLAB

#subject to tdrawpos {j in 1..nj}:
#	 cspline(t[1]*(sum {i in 1..nx} (x[i])), 0, nx,
#	 {i in 1..nx} x[i],
#	 {i in 1..nx-1} knots[i,j], vi, vf, ai, af) <= 0;
#subject to tdrawvel {j in 1..nj}:
#	 cspline(t[1]*(sum {i in 1..nx} (x[i])), 1, nx,
#	 {i in 1..nx} x[i],
#	 {i in 1..nx-1} knots[i,j], vi, vf, ai, af) <= 0;
#subject to tdrawacc {j in 1..nj}:
#	 cspline(t[1]*(sum {i in 1..nx} (x[i])), 2, nx,
#	 {i in 1..nx} x[i],
#	 {i in 1..nx-1} knots[i,j], vi, vf, ai, af) <= 0;
#subject to tdrawjerk {j in 1..nj}:
#	 cspline(t[1]*(sum {i in 1..nx} (x[i])), 3, nx,
#	 {i in 1..nx} x[i],
#	 {i in 1..nx-1} knots[i,j], vi, vf, ai, af) <= 0;


subject to tvel {j in 1..nj}:
	-limits[1,j] <= cspline(t[1]*(sum {i in 1..nx} (x[i])), 1, nx,
	 {i in 1..nx} x[i],
	 {i in 1..nx-1} knots[i,j], vi, vf, ai, af) <= limits[1,j];

subject to tacc {j in 1..nj}:
	-limits[2,j] <= cspline(t[1]*(sum {i in 1..nx} (x[i])), 2, nx,
	 {i in 1..nx} x[i],
	 {i in 1..nx-1} knots[i,j], vi, vf, ai, af) <= limits[2,j];

subject to tjerk {j in 1..nj}:
	-limits[3,j] <= cspline(t[1]*(sum {i in 1..nx} (x[i])), 3, nx,
	 {i in 1..nx} x[i],
	 {i in 1..nx-1} knots[i,j], vi, vf, ai, af) <= limits[3,j];


subject to bounds {i in 1..nx}:
	x[i] >= 0.01;
subject to tbounds {i in 1..nt}:
	0 <= t[i] <= 1;

data;

param nx := 9;
param nt := 1;

# Initial and final velocity and acceleration
param vi := 0;
param vf := 0;
param ai := 0;
param af := 0;

# Initial time
param it := 0;

# initial guess
param xinit:=
	1 3.607
	2 3.607
	3 2.878
	4 4.275
	5 5.612
	6 2.915
	7 5.879
	8 1.336
	9 1.336;
# solution by Lin et al. as initial guess
#param xinit:=
#	1 1.131
#	2 2.004
#	3 2.068
#	4 2.016
#	5 2.714
#	6 1.973
#	7 3.807
#	8 1.971
#	9 0.767;
# NSIPS solution, disc_hett method with default options
#param xinit:=
#	1  1.12515
#	2  2.03952
#	3  1.63594
#	4  2.15802
#	5  2.0466
#	6  2.51083
#	7  3.7812
#	8  1.83145
#	9  0.803105;


param tinit:=
	1 0;

# Via points in degrees
param nj  := 6;

param knots :
	1	2	3	4	5	6 :=
1	10	15	45	5	10	6
2	60	25	180	20	30	40
3	75	30	200	60	-40	80
4	130	-45	120	110	-60	70
5	110	-55	15	20	10	-10
6	100	-70	-10	60	50	10
7	-10	-10	100	-100	-40	30
8	-50	10	50	-30	10	20;

# Bound limits
param limits:
	1	2	3	4	5	6 :=
1	100	95	100	150	130	110
2	45	40	75	70	90	80
3	60	60	55	70	75	70;


# solve problem
# don't forget to write .col and .row files
option nsips_auxfiles rc;
option solver nsips;
solve;

##################
# Solution found #
##################
printf "Solution found\n";
display x;
display fx;

################################################
# solution of problem presented by the authors #
################################################
printf "Original Solution\n";
printf "x= %lf, %lf, %lf, %lf, %lf, %lf, %lf, %lf, %lf\n", 1.131, 2.004, 2.068,
	2.016, 2.714, 1.973, 3.807, 1.971, 0.767;
printf "fx= %lf\n", 18.451;

