###############################################################################
# Objective: Linear
# Constraints: General
###############################################################################
#
#   As described by C.G. Lo Bianco and A. Piazzi
#	in "A Semi-infinite optimization approach to optimal spline trajectory
#           planning of mechanical manipulators"
#    M.A. Goberna and M.A. Lopez (eds.), "Semi-Infinite Programming",
#           pp. 271-297, Kluwer Academic Publishers.
#
#   Mechanical arm with two revolute joints. Minimum time planning under torque
#    and velocities constraints
#
#    Coded by A. Ismael F. Vaz 28/03/2003  aivaz@dps.uminho.pt
#    University of Minho, Portugal
###############################################################################
function cspline;

# Gravity
param grav := 9.81;

# number of joints
param nj;

# Initial and final velocity and acceleration for each joint
param vi{1..nj};
param vf{1..nj};
param ai{1..nj};
param af{1..nj};

# Bound limits, velocity and torque
param limits{1..2,1..nj};

# Number of cubic sub-splines
# Number of finite variables
param nx;

# Number of spline parameters
# Number of infinite variables
param nt;

# Via points in joint space in rad
param knots{1..nx-1,1..nj};

# initial guess
param xinit{1..nx};
param tinit{1..nt};

# x variables
var x {i in 1..nx} := xinit[i];
# t variable
var t {j in 1..nt} := tinit[j];


# Robot parameters
param l1;
param l2;
param m1;
param m2;

# Cubic splines for dynamic equation
var q1   = cspline(t[1]*(sum {i in 1..nx} (x[i])), 0, nx, {i in 1..nx} x[i],
	 {i in 1..nx-1} knots[i,1], vi[1], vf[1], ai[1], af[1]);
var q2   = cspline(t[1]*(sum {i in 1..nx} (x[i])), 0, nx, {i in 1..nx} x[i],
	 {i in 1..nx-1} knots[i,2], vi[2], vf[2], ai[2], af[2]);
var q1d1 = cspline(t[1]*(sum {i in 1..nx} (x[i])), 1, nx, {i in 1..nx} x[i],
	 {i in 1..nx-1} knots[i,1], vi[1], vf[1], ai[1], af[1]);
var q2d1 = cspline(t[1]*(sum {i in 1..nx} (x[i])), 1, nx, {i in 1..nx} x[i],
	 {i in 1..nx-1} knots[i,2], vi[2], vf[2], ai[2], af[2]);
var q1d2 = cspline(t[1]*(sum {i in 1..nx} (x[i])), 2, nx, {i in 1..nx} x[i],
	 {i in 1..nx-1} knots[i,1], vi[1], vf[1], ai[1], af[1]);
var q2d2 = cspline(t[1]*(sum {i in 1..nx} (x[i])), 2, nx, {i in 1..nx} x[i],
	 {i in 1..nx-1} knots[i,2], vi[2], vf[2], ai[2], af[2]);

# Dynamic equation for torque
var u1 = m2*l2*l2*(q1d2+q2d2)+m2*l1*l2*cos(q2)*(2*q1d2+q2d2)+
	(m1+m2)*l1*l1*q1d2-m2*l1*l2*sin(q2)*q2d1^2-
	2*m2*l1*l2*sin(q2)*q1d1*q2d1+m2*l2*grav*cos(q1+q2)+
	(m1+m2)*l1*grav*cos(q1);
var u2 = m2*l1*l2*cos(q2)*q1d2+m2*l1*l2*sin(q2)*q1d1^2+
	m2*l2*grav*cos(q1+q2)+m2*l2*l2*(q1d2+q2d2);

# Dynamic equation for linear velocity
var v = (l1*l1*q1d1^2+l2*l2*(q1d1+q2d1)^2
	  +2*l1*l2*q1d1*(q1d1+q2d1)*cos(q2)); # Square root

# Dynamic equation for angular velocity
var w = q1d1+q2d1;


minimize fx:
	sum {i in 1..nx} (x[i]);


# These constraint can be used to display the cubic splines under MATLAB

#subject to tdrawpos {j in 1..nj}:
#	 cspline(t[1]*(sum {i in 1..nx} (x[i])), 0, nx,
#	 {i in 1..nx} x[i],
#	 {i in 1..nx-1} knots[i,j], vi[j], vf[j], ai[j], af[j]) <= 0;
#subject to tdrawvel {j in 1..nj}:
#	 cspline(t[1]*(sum {i in 1..nx} (x[i])), 1, nx,
#	 {i in 1..nx} x[i],
#	 {i in 1..nx-1} knots[i,j], vi[j], vf[j], ai[j], af[j]) <= 0;
#subject to tdrawacc {j in 1..nj}:
#	 cspline(t[1]*(sum {i in 1..nx} (x[i])), 2, nx,
#	 {i in 1..nx} x[i],
#	 {i in 1..nx-1} knots[i,j], vi[j], vf[j], ai[j], af[j]) <= 0;
#subject to tdrawjerk {j in 1..nj}:
#	 cspline(t[1]*(sum {i in 1..nx} (x[i])), 3, nx,
#	 {i in 1..nx} x[i],
#	 {i in 1..nx-1} knots[i,j], vi[j], vf[j], ai[j], af[j]) <= 0;
#subject to tdrawtorque1:
#	u1 <=0;
#subject to tdrawtorque2:
#	u2 <=0;
#subject to tdrawlvel:
#	v <= 0;
#subject to tdrawavel:
#	w <= 0;


subject to torque1:
	-limits[1,1] <= u1 <= limits[1,1];

subject to torque2:
	-limits[1,2] <= u2 <= limits[1,2];

subject to tlinearvel:
	-limits[2,1] <= v <= limits[2,1];

subject to tangularvel:
	-limits[2,2] <= w <= limits[2,2];


subject to bounds {i in 1..nx}:
	x[i] >= 0.01;
subject to tbounds {i in 1..nt}:
	0 <= t[i] <= 1;

data;

param nx := 11;
param nt := 1;

# Initial and final velocity and acceleration
param vi :=
	1 0
	2 0;
param vf :=
	1 0
	2 0;
param ai := 
	1  0
	2  0;
param af := 
	1 0
	2 0;

# Robot parameters
param l1 := 1.0; 
param l2 := 0.5;
param m1 := 15;
param m2 := 7;



# initial guess
param xinit:=
	1  0.5
	2  0.5	# extra space for extra freedom
	3  0.5
	4  0.5
	5  0.5
	6  0.5
	7  0.5
	8  0.5
	9  0.5
	10 0.5	# extra space for extra freedom
	11 0.5;


# NSIPS solution, disc_hett method with default options
#param xinit:=
#	 1  0.01
#	 2  0.348599
#	 3  0.156699
#	 4  0.150559
#	 5  0.154683
#	 6  0.13814
#	 7  0.191483
#	 8  0.391619
#	 9  0.106903
#	10  0.0226164
#	11  0.385888;



param tinit:=
	1 0;

# Via points in degrees
param nj  := 2;

param knots :
	1	2 :=
1	0.0000	-1.5708
2	0.1253	-1.6804
3	0.2517	-1.7594
4	0.3789	-1.8074
5	0.5054	-1.8235
6	0.5837	-1.7087
7	0.6119	-1.4581
8	0.4263	-1.1040
9	0.3903	-1.1124
10	0.3526	-1.1152;


# Bound limits
param limits:
	1	2 :=
1	260	50  # torque
2	0.7	2.25;# linear and angular velocity


# solve problem
# don't forget to write .col and .row files
option nsips_auxfiles rc;
option solver nsips;
solve;

##################
# Solution found #
##################
printf "Solution found\n";
display x;
display fx;

################################################
# solution of problem presented by the authors #
################################################
printf "Original Solution\n";
printf "h= %lf, %lf, %lf, %lf, %lf, %lf, %lf, %lf, %lf, %lf, %lf\n", 0.02000, 0.36429, 0.18419, 0.18386, 0.18423, 0.16735, 0.22310, 0.36539, 0.09945, 0.23818, 0.02005;
printf "Total time= %lf\n", 2.05009;

